###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from unittest import TestCase, mock
from everysk.core.compress import compress, decompress, zip_directory_to_str
from everysk.core.datetime import DateTime, Date
from everysk.core import serialize

class CompressTestCase(TestCase):
    @mock.patch.object(serialize, 'dumps')
    def test_compress_serialize_none(self, dumps: mock.MagicMock):
        result = compress('a', serialize=None)
        self.assertEqual(result, b'x\x9cK\x04\x00\x00b\x00b')
        dumps.assert_not_called()

    @mock.patch.object(serialize, 'loads')
    def test_decompress_serialize_none(self, loads: mock.MagicMock):
        result = decompress(b'x\x9cK\x04\x00\x00b\x00b', serialize=None)
        self.assertEqual(result, b'a')
        loads.assert_not_called()

    def test_compress_protocol_invalid(self):
        with self.assertRaisesRegex(ValueError, "Unsupported compression protocol 'invalid'. Use 'zlib' or 'gzip'."):
            compress('string', protocol='invalid', serialize='json')

    def test_decompress_protocol_invalid(self):
        with self.assertRaisesRegex(ValueError, "Unsupported decompression protocol 'invalid'. Use 'zlib' or 'gzip'."):
            decompress('string', protocol='invalid', serialize='json')


class CompressZlibJsonTestCase(TestCase):

    def test_compress(self):
        string = 'aa aa aa aa aa aa aa aa'
        self.assertEqual(compress(string, protocol='zlib', serialize='json'), b'x\x9cSJLT\xc0\x86\x94\x00]\xbd\x075')

    def test_decompress(self):
        string = b'x\x9cSJLT\xc0\x86\x94\x00]\xbd\x075'
        self.assertEqual(decompress(string, protocol='zlib', serialize='json'), 'aa aa aa aa aa aa aa aa')

    def test_undefined(self):
        obj_compressed = compress(Undefined, protocol='zlib', serialize='json')
        self.assertEqual(decompress(obj_compressed, protocol='zlib', serialize='json'), Undefined)

    def test_complex_obj(self):
        obj = [
            '2024-01-01',
            {'datetime': DateTime.now()},
            {'a': 1, 'b': True},
            'Test',
            1.1,
            ['1', '2', '3', Date.today()],
            Undefined,
            {'undefined': Undefined}
        ]
        obj_compressed = compress(obj, protocol='zlib', serialize='json')
        self.assertListEqual(decompress(obj_compressed, protocol='zlib', serialize='json'), obj)

    def test_date_format(self):
        obj = Date.today()
        with mock.patch.dict('os.environ', {'SERIALIZE_DATE_FORMAT':'%Y%m%d'}):
            obj_compressed = compress(obj, protocol='zlib', serialize='json')
            self.assertEqual(decompress(obj_compressed, protocol='zlib', serialize='json'), obj.strftime('%Y%m%d'))

    def test_datetime_format(self):
        obj = DateTime.now()
        with mock.patch.dict('os.environ', {'SERIALIZE_DATETIME_FORMAT':'%Y%m%d %HH:%M:%S'}):
            obj_compressed = compress(obj, protocol='zlib', serialize='json')
            self.assertEqual(decompress(obj_compressed, protocol='zlib', serialize='json'), obj.strftime('%Y%m%d %HH:%M:%S'))


class CompressZlibPickleTestCase(TestCase):

    def setUp(self) -> None:
        self.str_repr = b'x\x9ck`\x99*\xcd\x00\x01=\xe2\x89\x89\n\xd8\xd0\x14=\x00\x9f\xef\t\x8a'

    def test_compress(self):
        string = 'aa aa aa aa aa aa aa aa'
        self.assertEqual(compress(string, protocol='zlib', serialize='pickle'), self.str_repr)

    def test_decompress(self):
        self.assertEqual(decompress(self.str_repr, protocol='zlib', serialize='pickle'), 'aa aa aa aa aa aa aa aa')

    def test_undefined(self):
        obj_compressed = compress(Undefined, protocol='zlib', serialize='pickle')
        self.assertEqual(decompress(obj_compressed, protocol='zlib', serialize='pickle'), Undefined)

    def test_complex_obj(self):
        obj = [
            '2024-01-01',
            {'datetime': DateTime.now()},
            {'a': 1, 'b': True},
            'Test',
            1.1,
            ['1', '2', '3', Date.today()],
            Undefined,
            {'undefined': Undefined}
        ]
        obj_compressed = compress(obj, protocol='zlib', serialize='pickle')
        self.assertListEqual(decompress(obj_compressed, protocol='zlib', serialize='pickle'), obj)
###############################################################################
#   File Handling Test Case Implementation
###############################################################################
class FileHandlingTestCase(TestCase):

    @mock.patch('os.walk')
    @mock.patch('zipfile.ZipFile.write')
    def test_zip_directory_to_str_returns_expected_data(self, mock_walk, mock_write):
        expected_base64 = 'UEsFBgAAAAAAAAAAAAAAAAAAAAAAAA=='
        mock_walk.return_value = [
            ('/engine', ('portfolio'), ('tests')),
            ('/entities/file', (), ('base', 'settings'))
        ]
        mock_write.return_value = ['123', '456']
        result = zip_directory_to_str('engine/portfolio/tests', 'zip_root_folder')

        self.assertTrue(result.isascii())
        self.assertTrue(mock_write.called)
        self.assertTrue(mock_walk.called)
        self.assertIsInstance(result, str)
        self.assertEqual(result, expected_base64)

    @mock.patch('os.walk')
    def test_zip_directory_to_str_with_invalid_paths(self, mock_walk):
        with self.assertRaises(FileNotFoundError) as context:
            mock_walk.return_value = [
            ('/engine', ('portfolio'), ('tests')),
            ('/entities/file', (), ('base', 'settings'))
        ]
            zip_directory_to_str('invalid/path', 'invalid_root_folder')

        self.assertEqual(str(context.exception), "[Errno 2] No such file or directory: '/engine/t'")

    def test_date_format(self):
        obj = Date.today()
        with mock.patch.dict('os.environ', {'SERIALIZE_DATE_FORMAT':'%Y%m%d'}):
            obj_compressed = compress(obj, protocol='zlib', serialize='pickle')
            self.assertEqual(decompress(obj_compressed, protocol='zlib', serialize='pickle'), obj)

    def test_datetime_format(self):
        obj = DateTime.now()
        with mock.patch.dict('os.environ', {'SERIALIZE_DATETIME_FORMAT':'%Y%m%d %HH:%M:%S'}):
            obj_compressed = compress(obj, protocol='zlib', serialize='pickle')
            self.assertEqual(decompress(obj_compressed, protocol='zlib', serialize='pickle'), obj)


class CompressGzipJsonTestCase(TestCase):

    def test_undefined(self):
        obj_compressed = compress(Undefined, protocol='gzip', serialize='json')
        self.assertEqual(decompress(obj_compressed, protocol='gzip', serialize='json'), Undefined)

    def test_complex_obj(self):
        obj = [
            '2024-01-01',
            {'datetime': DateTime.now()},
            {'a': 1, 'b': True},
            'Test',
            1.1,
            ['1', '2', '3', Date.today()],
            Undefined,
            {'undefined': Undefined}
        ]
        obj_compressed = compress(obj, protocol='gzip', serialize='json')
        self.assertListEqual(decompress(obj_compressed, protocol='gzip', serialize='json'), obj)

    def test_date_format(self):
        obj = Date.today()
        with mock.patch.dict('os.environ', {'SERIALIZE_DATE_FORMAT':'%Y%m%d'}):
            obj_compressed = compress(obj, protocol='gzip', serialize='json')
            self.assertEqual(decompress(obj_compressed, protocol='gzip', serialize='json'), obj.strftime('%Y%m%d'))

    def test_datetime_format(self):
        obj = DateTime.now()
        with mock.patch.dict('os.environ', {'SERIALIZE_DATETIME_FORMAT':'%Y%m%d %HH:%M:%S'}):
            obj_compressed = compress(obj, protocol='gzip', serialize='json')
            self.assertEqual(decompress(obj_compressed, protocol='gzip', serialize='json'), obj.strftime('%Y%m%d %HH:%M:%S'))


class CompressGzipPickleTestCase(TestCase):

    def test_undefined(self):
        obj_compressed = compress(Undefined, protocol='gzip', serialize='pickle')
        self.assertEqual(decompress(obj_compressed, protocol='gzip', serialize='pickle'), Undefined)

    def test_complex_obj(self):
        obj = [
            '2024-01-01',
            {'datetime': DateTime.now()},
            {'a': 1, 'b': True},
            'Test',
            1.1,
            ['1', '2', '3', Date.today()],
            Undefined,
            {'undefined': Undefined}
        ]
        obj_compressed = compress(obj, protocol='gzip', serialize='pickle')
        self.assertListEqual(decompress(obj_compressed, protocol='gzip', serialize='pickle'), obj)

    def test_date_format(self):
        obj = Date.today()
        with mock.patch.dict('os.environ', {'SERIALIZE_DATE_FORMAT':'%Y%m%d'}):
            obj_compressed = compress(obj, protocol='gzip', serialize='pickle')
            self.assertEqual(decompress(obj_compressed, protocol='gzip', serialize='pickle'), obj)

    def test_datetime_format(self):
        obj = DateTime.now()
        with mock.patch.dict('os.environ', {'SERIALIZE_DATETIME_FORMAT':'%Y%m%d %HH:%M:%S'}):
            obj_compressed = compress(obj, protocol='gzip', serialize='pickle')
            self.assertEqual(decompress(obj_compressed, protocol='gzip', serialize='pickle'), obj)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from unittest import TestCase, mock
from unittest.mock import MagicMock
from unittest import mock

from everysk.config import settings
from everysk.core.http import HttpSDKPOSTConnection
from everysk.core.exceptions import FieldValueError

from everysk.sdk.entities.script import Script
from everysk.sdk.entities.portfolio.base import Portfolio


###############################################################################
#   Script TestCase Implementation
###############################################################################
class ScriptTestCase(TestCase):
    def setUp(self):
        self.mock_klass = MagicMock()
        self.script = Script(Portfolio)
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()

    def test_script_query_none_user_input(self):
        result = Portfolio.script.fetch(None, 'someVariant', 'someWorkspace')
        self.assertIsNone(result)

    def test_script_query_previous_workers_without_id(self):
        user_input = {"someField": "someValue"}
        result = Portfolio.script.fetch(user_input, 'previousWorkers', 'someWorkspace')
        self.assertIsInstance(result, Portfolio)

    def test_script_query_tag_latest_variant(self):
        expected_data = '{"class_name": "Script", "method_name": "fetch", "self_obj": {"_klass": "Portfolio", "__class_path__": "everysk.sdk.entities.script.Script"}, "params": {"user_input": {"tags": "someTag"}, "variant": "tagLatest", "workspace": "someWorkspace"}}'

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            Portfolio.script.fetch({"tags": "someTag"}, 'tagLatest', 'someWorkspace')

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )

    def test_script_query_when_response_is_a_list(self):
        with mock.patch('requests.post') as mock_post, \
            mock.patch('everysk.core.http.dumps') as mock_dumps:
            mock_dumps.return_value = 'test'
            mock_post.return_value.content = '[{}, {}]'
            mock_post.return_value.status_code = 200
            result = Portfolio.script.fetch(['tag1', 'tag2'], 'tagList', 'someWorkspace')

        mock_dumps.assert_called_once_with(
            {
                'class_name': 'Script',
                'method_name': 'fetch',
                'self_obj': Portfolio.script,
                'params': {
                    'user_input': ['tag1', 'tag2'],
                    'variant': 'tagList',
                    'workspace': 'someWorkspace'
                }
            },
            use_undefined=True
        )
        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=mock_dumps.return_value
        )
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 2)
        self.assertIsInstance(result[0], Portfolio)
        self.assertIsInstance(result[1], Portfolio)

    def test_persist_transient(self):
        entity = Portfolio(name='test', workspace='someWorkspace')

        with mock.patch('requests.post') as mock_post, \
            mock.patch('everysk.core.http.dumps') as mock_dumps:
            mock_dumps.return_value = 'test'
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = Portfolio.script.persist(entity, 'transient')

        mock_dumps.assert_called_once_with(
            {
                'class_name': 'Script',
                'method_name': 'persist',
                'self_obj': Portfolio.script,
                'params': {
                    'entity': entity,
                    'persist': 'transient',
                    'consistency_check': False
                }
            },
            use_undefined=True
        )
        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=mock_dumps.return_value
        )
        self.assertIsInstance(result, Portfolio)

    def test_init_(self):
        # pylint: disable=protected-access
        script = Portfolio.script
        self.assertEqual(script._klass, Portfolio)

        script = Script(Portfolio)
        self.assertEqual(script._klass, Portfolio)

        script = Script(_klass=Portfolio)
        self.assertEqual(script._klass, Portfolio)

        script = Script(_klass='Portfolio')
        self.assertEqual(script._klass, Portfolio)

        self.assertRaisesRegex(
            FieldValueError,
            "The _klass value 'Foo' must be a class or a string with the class name",
            Script,
            _klass='Foo'
        )



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from unittest import TestCase, mock

from everysk.config import settings

from everysk.sdk.engines.cache import UserCache

from everysk.core.http import HttpSDKPOSTConnection

###############################################################################
#   Cache Test Case Implementation
###############################################################################
class CacheTestCase(TestCase):
    def setUp(self) -> None:
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()
        self.expected_error_message = 'Invalid time value. The time value should be an integer greater than 0 and less than or equal to the default expiration time.'
        return super().setUp()

    ###############################################################################
    #   Get Method Test Case Implementation
    ###############################################################################
    def test_get_method_assert_is_called(self):
        expected_data = '{"class_name": "UserCache", "method_name": "get", "self_obj": {"__class_path__": "everysk.sdk.engines.cache.UserCache"}, "params": {"key": "key", "prefix": ""}}'
        expected_result = {}
        cache = UserCache()

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = cache.get("key")

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )
        self.assertEqual(result, expected_result)

    ###############################################################################
    #   Get Multi Method Test Case Implementation
    ###############################################################################
    def test_get_multi_method_assert_is_called(self):
        expected_data = '{"class_name": "UserCache", "method_name": "get_multi", "self_obj": {"__class_path__": "everysk.sdk.engines.cache.UserCache"}, "params": {"keys": ["key1", "key2"], "prefix": ""}}'
        expected_result = {}
        cache = UserCache()

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = cache.get_multi(['key1', 'key2'])

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )
        self.assertEqual(result, expected_result)

    ###############################################################################
    #   Set Method Test Case Implementation
    ###############################################################################
    def test_set_assert_is_called(self):
        expected_data = '{"class_name": "UserCache", "method_name": "set", "self_obj": {"__class_path__": "everysk.sdk.engines.cache.UserCache"}, "params": {"key": "key", "data": "value", "time": 14400, "prefix": ""}}'
        cache = UserCache()
        expected_result = {}

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = cache.set(key="key", data="value")

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )
        self.assertEqual(result, expected_result)

    def test_set_raises_error_with_invalid_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().set(key="key", data="value", time='123')

        self.assertEqual(self.expected_error_message, str(context.exception))

    def test_set_raises_error_with_negative_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().set(key="key", data="value", time=-2)

        self.assertEqual(self.expected_error_message, str(context.exception))

    def test_set_raises_error_with_greater_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().set(key="key", data="value", time=settings.ENGINES_CACHE_EXECUTION_EXPIRATION_TIME + 10)

        self.assertEqual(self.expected_error_message, str(context.exception))

    ###############################################################################
    #   Set Multi Method Test Case Implementation
    ###############################################################################
    def test_set_multi_assert_is_called(self):
        expected_data = '{"class_name": "UserCache", "method_name": "set_multi", "self_obj": {"__class_path__": "everysk.sdk.engines.cache.UserCache"}, "params": {"data_dict": {"key1": "value1", "key2": "value2"}, "time": 14400, "prefix": ""}}'
        cache = UserCache()
        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            cache.set_multi({'key1': 'value1', 'key2': 'value2'})

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )

    def test_set_multi_raises_error_with_invalid_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().set_multi(data_dict={'key1': 'value1', 'key2': 'value2'}, time='123')

        self.assertEqual(self.expected_error_message, str(context.exception))

    def test_set_multi_raises_error_with_negative_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().set_multi(data_dict={'key1': 'value1', 'key2': 'value2'}, time=-1)

        self.assertEqual(self.expected_error_message, str(context.exception))

    def test_set_multi_raises_error_with_greater_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().set_multi(data_dict={'key1': 'value1', 'key2': 'value2'}, time=settings.ENGINES_CACHE_EXECUTION_EXPIRATION_TIME + 10)

        self.assertEqual(self.expected_error_message, str(context.exception))

    ###############################################################################
    #   Incr Method Test Case Implementation
    ###############################################################################
    def test_incr_assert_is_called(self):
        expected_data = '{"class_name": "UserCache", "method_name": "incr", "self_obj": {"__class_path__": "everysk.sdk.engines.cache.UserCache"}, "params": {"key": "key", "delta": 1, "initial_value": 0, "time": 14400}}'
        cache = UserCache()
        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            cache.incr(key='key', delta=1, initial_value=0)

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )

    def test_incr_raises_error_with_invalid_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().incr(key=1, delta=1, initial_value=0, time='123')

        self.assertEqual(self.expected_error_message, str(context.exception))

    def test_incr_raises_error_with_negative_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().incr(key=1, delta=1, initial_value=0, time=-1)

        self.assertEqual(self.expected_error_message, str(context.exception))

    def test_incr_raises_error_with_greater_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().incr(key=1, delta=1, initial_value=0, time=settings.ENGINES_CACHE_EXECUTION_EXPIRATION_TIME + 10)

        self.assertEqual(self.expected_error_message, str(context.exception))

    ###############################################################################
    #   Decr Method Test Case Implementation
    ###############################################################################
    def test_decr_assert_is_called(self):
        expected_data = '{"class_name": "UserCache", "method_name": "decr", "self_obj": {"__class_path__": "everysk.sdk.engines.cache.UserCache"}, "params": {"key": "key", "delta": 1, "initial_value": 0, "time": 14400}}'
        cache = UserCache()
        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            cache.decr(key='key', delta=1, initial_value=0)

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )

    def test_decr_raises_error_with_invalid_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().decr(key='key', delta=1, initial_value=0, time='123')

        self.assertEqual(self.expected_error_message, str(context.exception))

    def test_decr_raises_error_with_negative_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().decr(key='key', delta=1, initial_value=0, time=-1)

        self.assertEqual(self.expected_error_message, str(context.exception))

    def test_decr_raises_error_with_greater_time(self):
        with self.assertRaises(ValueError) as context:
            UserCache().decr(key='key', delta=1, initial_value=0, time=settings.ENGINES_CACHE_EXECUTION_EXPIRATION_TIME + 10)

        self.assertEqual(self.expected_error_message, str(context.exception))

    ###############################################################################
    #   Delete Method Test Case Implementation
    ###############################################################################
    def test_delete_assert_is_called(self):
        expected_data = '{"class_name": "UserCache", "method_name": "delete", "self_obj": {"__class_path__": "everysk.sdk.engines.cache.UserCache"}, "params": {"key": "key"}}'
        cache = UserCache()
        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            cache.delete(key='key')

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )

    ###############################################################################
    #   Delete Multi Method Test Case Implementation
    ###############################################################################
    def test_delete_multi_assert_is_called(self):
        expected_data = '{"class_name": "UserCache", "method_name": "delete_multi", "self_obj": {"__class_path__": "everysk.sdk.engines.cache.UserCache"}, "params": {"keys": ["key1", "key2"]}}'
        cache = UserCache()
        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            cache.delete_multi(keys=['key1', 'key2'])

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=invalid-name, protected-access
from os import environ
from unittest import TestCase, mock, skip
from everysk.core import firestore
from everysk.core.compress import compress
from everysk.config import settings
from everysk.core.datetime import Date, DateTime
from everysk.core.exceptions import FieldValueError
from everysk.core import redis as redis_module
from everysk.core.threads import ThreadPool


def _test_connection(Client: mock.MagicMock):
    doc = firestore.FirestoreClient(project_name='P01', database_name='D01')
    assert doc.connection == Client.return_value

@mock.patch.object(firestore.firestore, 'Client')
class FirestoreClientTestCase(TestCase):

    def setUp(self) -> None:
        firestore.FirestoreClient._connections = {}

    def test_connections(self, Client: mock.MagicMock):
        doc01 = firestore.FirestoreClient(project_name='P01', database_name='D01')
        self.assertEqual(doc01.project_name, 'P01')
        self.assertEqual(doc01.database_name, 'D01')
        self.assertEqual(doc01.connection, Client.return_value)
        Client.assert_called_once_with(project='P01', database='D01')

        Client.reset_mock()
        doc02 = firestore.FirestoreClient(project_name='P02', database_name='D02')
        self.assertEqual(doc02.project_name, 'P02')
        self.assertEqual(doc02.database_name, 'D02')
        self.assertEqual(doc02.connection, Client.return_value)
        Client.assert_called_once_with(project='P02', database='D02')

    def test_concurrent_connections(self, Client: mock.MagicMock):
        pool = ThreadPool(4)
        for i in range(0, 13): # pylint: disable=unused-variable
            pool.add(
                target=_test_connection,
                kwargs={'Client': Client}
            )
        pool.wait()
        Client.assert_called_once_with(project='P01', database='D01')

    @mock.patch.object(firestore, 'RedisLock')
    def test_connection_lock(self, RedisLock: mock.MagicMock, Client: mock.MagicMock):
        doc01 = firestore.FirestoreClient(project_name='P01', database_name='D01')
        doc01.connection # pylint: disable=pointless-statement
        Client.assert_called_once_with(project='P01', database='D01')
        RedisLock.assert_called_once_with(name='everysk-lib-firestore-lock-connection-P01-D01', timeout=600)
        RedisLock.return_value.acquire.assert_called_once_with(blocking=True)
        RedisLock.return_value.release.assert_called_once_with()

    @mock.patch.object(firestore, 'RedisLock')
    def test_connection_error(self, RedisLock: mock.MagicMock, Client: mock.MagicMock):
        Client.side_effect = AttributeError('Erro')
        doc01 = firestore.FirestoreClient(project_name='P01', database_name='D01')
        with self.assertRaisesRegex(AttributeError, 'Erro'):
            doc01.connection # pylint: disable=pointless-statement
        Client.assert_called_once_with(project='P01', database='D01')
        RedisLock.assert_called_once_with(name='everysk-lib-firestore-lock-connection-P01-D01', timeout=600)
        RedisLock.return_value.acquire.assert_called_once_with(blocking=True)
        RedisLock.return_value.release.assert_called_once_with()

    def test_init_default(self, Client: mock.MagicMock):
        client = firestore.FirestoreClient()
        self.assertEqual(client.database_name, firestore._DEFAULT_DATABASE)
        self.assertEqual(client.project_name, settings.EVERYSK_GOOGLE_CLOUD_PROJECT)
        Client.assert_not_called()

    def test_init_params(self, Client: mock.MagicMock):
        client = firestore.FirestoreClient(project_name='PROJECT', database_name='DATABASE')
        self.assertEqual(client.database_name, 'DATABASE')
        self.assertEqual(client.project_name, 'PROJECT')
        Client.assert_not_called()

    def test_get_collection(self, Client: mock.MagicMock):
        client = firestore.FirestoreClient()
        collection = client.get_collection('my-collection')
        self.assertEqual(collection, Client.return_value.collection.return_value)
        Client.return_value.collection.assert_called_once_with('my-collection')

    def test_environment_values(self, Client: mock.MagicMock):
        environ['EVERYSK_GOOGLE_CLOUD_PROJECT'] = 'PROJECT'
        client = firestore.FirestoreClient()
        self.assertEqual(client.project_name, 'PROJECT')
        Client.assert_not_called()
        del environ['EVERYSK_GOOGLE_CLOUD_PROJECT']


@mock.patch.object(firestore.firestore, 'Client')
class BaseDocumentConfigTestCase(TestCase):

    def setUp(self) -> None:
        firestore.FirestoreClient._connections = {}

    def test_client(self, Client: mock.MagicMock):
        obj = firestore.BaseDocumentConfig(collection_name='test-collection')
        self.assertIsInstance(obj.client, firestore.FirestoreClient)
        Client.assert_not_called()

    def test_collection(self, Client: mock.MagicMock):
        obj = firestore.BaseDocumentConfig(collection_name='test-collection')
        self.assertEqual(obj.collection, Client.return_value.collection.return_value)
        Client.return_value.collection.assert_called_once_with('test-collection')

    def test_collection_empty(self, Client: mock.MagicMock):
        obj = firestore.BaseDocumentConfig()
        with self.assertRaisesRegex(AttributeError, 'The collection_name is empty.'):
            obj.collection # pylint: disable=pointless-statement
        Client.return_value.collection.assert_not_called()

    def test_excluded_keys(self, Client: mock.MagicMock):
        obj = firestore.BaseDocumentConfig(collection_name='test-collection')
        self.assertListEqual(obj.excluded_keys, [])
        Client.assert_not_called()


class FakeObject:
    """ Used to test pickle and unpickle """
    value: int = 1

    def __eq__(self, __value: object) -> bool:
        return self.__class__ == __value.__class__ and self.value == __value.value


@mock.patch.object(firestore.firestore, 'Client')
class DocumentTestCase(TestCase):

    def setUp(self) -> None:
        firestore.FirestoreClient._connections = {}
        firestore.Document.config._client = None

    def test_config(self, Client: mock.MagicMock):
        doc = firestore.Document()
        self.assertIsInstance(doc.config, firestore.BaseDocumentConfig)
        Client.assert_not_called()

    @mock.patch.object(firestore.DateTime, 'now', return_value='2022-01-01T10:00:00+00:00')
    def test_init(self, now: mock.MagicMock, Client: mock.MagicMock):
        doc = firestore.Document(key='value', firestore_id='my-id')
        self.assertEqual(doc.firestore_id, 'my-id')
        self.assertEqual(doc.created_at, DateTime.fromisoformat('2022-01-01T10:00:00+00:00'))
        self.assertEqual(doc.key, 'value')
        now.assert_called_once_with()
        Client.assert_not_called()

    def test_parse_in(self, Client: mock.MagicMock):
        doc = firestore.Document(
            dct={'a': 1},
            lst=[1, 2, 3],
            date='2022-01-01',
            datetime='2022-01-01T10:00:00+00:00',
            obj=compress(FakeObject(), serialize='pickle'),
            byte=b'Text'
        )
        self.assertDictEqual(doc.dct, {'a': 1})
        self.assertListEqual(doc.lst, [1, 2, 3])
        self.assertEqual(doc.date, Date.fromisoformat('2022-01-01'))
        self.assertEqual(doc.datetime, DateTime.fromisoformat('2022-01-01T10:00:00+00:00'))
        self.assertEqual(doc.obj, FakeObject())
        self.assertEqual(doc.byte, b'Text')
        Client.assert_not_called()

    def test_parser_out(self, Client: mock.MagicMock):
        doc = firestore.Document()
        self.assertDictEqual(doc._parser_out({'a': 1}), {'a': 1})
        self.assertListEqual(doc._parser_out([1, 2, 3]), [1, 2, 3])
        self.assertEqual(doc._parser_out(Date.fromisoformat('2022-01-01')), '2022-01-01')
        self.assertEqual(
            doc._parser_out(DateTime.fromisoformat('2022-01-01T10:00:00+00:00')), '2022-01-01T10:00:00+00:00'
        )
        obj = FakeObject()
        self.assertEqual(doc._parser_out(obj), compress(obj, serialize='pickle'))
        Client.assert_not_called()

    def test_loads(self, Client: mock.MagicMock):
        firestore.Document.config.collection_name = 'my-collection'
        Client.return_value.collection.return_value.where.return_value.order_by.return_value.limit.return_value.get.return_value = [
            firestore.Document(firestore_id='id01', type='1', mic='XYSX', created_at='2023-04-03T00:00:00+00:00')
        ]
        doc = firestore.Document.loads(field='firestore_id', condition='==', value='my-id')
        Client.return_value.collection.return_value.where.assert_called_once_with('firestore_id', '==', 'my-id')
        Client.return_value.collection.return_value.where.return_value.order_by.assert_called_once_with('firestore_id')
        self.assertListEqual(
            doc,
            [firestore.Document(**{
                'created_at': DateTime.fromisoformat('2023-04-03T00:00:00+00:00'),
                'firestore_id': 'id01',
                'mic': 'XYSX',
                'type': '1',
                'updated_at': None
            })]
        )
        firestore.Document.config.collection_name = None

    def test_get_firestore_id(self, Client: mock.MagicMock):
        doc = firestore.Document()
        self.assertNotEqual(doc.get_firestore_id(), doc.get_firestore_id())
        doc = firestore.Document(firestore_id='Test')
        self.assertEqual(doc.get_firestore_id(), 'Test')
        Client.assert_not_called()

    def test_load(self, Client: mock.MagicMock):
        firestore.Document.config.collection_name = 'my-collection'
        Client.return_value.collection.return_value.document.return_value.get.return_value.exists = True
        Client.return_value.collection.return_value.document.return_value.get.return_value.to_dict.return_value = {
            'created_at': '2023-04-03T00:00:00+00:00',
            'dct': {'a': 1},
            'lst': [1, 2, 3],
            'date': '2022-01-01',
            'datetime': '2022-01-01T10:00:00+00:00',
            'obj': compress(FakeObject(), serialize='pickle'),
            'byte': b'Text'
        }
        doc = firestore.Document(firestore_id='my-id')
        doc.load()
        Client.return_value.collection.return_value.document.assert_called_once_with('my-id')
        self.assertDictEqual(
            doc,
            {
                'byte': b'Text',
                'created_at': DateTime.fromisoformat('2023-04-03T00:00:00+00:00'),
                'date': Date(2022, 1, 1),
                'datetime': DateTime.fromisoformat('2022-01-01T10:00:00+00:00'),
                'dct': {'a': 1},
                'firestore_id': 'my-id',
                'lst': [1, 2, 3],
                'obj': FakeObject(),
                'updated_at': None
            }
        )
        firestore.Document.config.collection_name = None

    @mock.patch.object(firestore.DateTime, 'now', return_value='2023-10-11T00:00:00+00:00')
    def test_load_no_exists(self, now: mock.MagicMock, Client: mock.MagicMock):
        firestore.Document.config.collection_name = 'my-collection'
        Client.return_value.collection.return_value.document.return_value.get.return_value.exists = False
        Client.return_value.collection.return_value.document.return_value.get.return_value.to_dict.return_value = {
            'created_at': '2023-04-03T00:00:00+00:00',
            'dct': {'a': 1},
            'lst': [1, 2, 3],
            'date': '2022-01-01',
            'datetime': '2022-01-01T10:00:00+00:00',
            'obj': compress(FakeObject(), serialize='pickle'),
            'byte': b'Text'
        }
        doc = firestore.Document(firestore_id='my-id')
        doc.load()
        Client.return_value.collection.return_value.document.assert_called_once_with('my-id')
        self.assertDictEqual(
            doc,
            {
                'created_at': DateTime.fromisoformat('2023-10-11T00:00:00+00:00'),
                'firestore_id': 'my-id',
                'updated_at': None
            }
        )
        firestore.Document.config.collection_name = None
        now.assert_called_once_with()

    @skip('Error on the test, need to fix it.')
    @mock.patch.object(firestore.DateTime, 'now', return_value='2023-10-11T00:00:00+00:00')
    def test_save(self, now: mock.MagicMock, Client: mock.MagicMock):
        firestore.Document.config.collection_name = 'my-collection'
        doc = firestore.Document(firestore_id='my-id', var='1', created_at='2023-04-03T00:00:00+00:00')
        doc.save()
        Client.return_value.collection.return_value.document.assert_called_once_with('my-id')
        Client.return_value.collection.return_value.document.return_value.set.assert_called_once_with(
            document_data={'created_at': '2023-04-03T00:00:00+00:00', 'firestore_id': 'my-id', 'updated_at': '2023-10-11T00:00:00+00:00', 'var': '1'},
            merge=True,
            timeout=60.0
        )
        now.assert_called_once_with()
        firestore.Document.config.collection_name = None

    def test_to_dict(self, Client: mock.MagicMock):
        obj = FakeObject()
        doc = firestore.Document(
            byte=b'Text',
            created_at=DateTime.fromisoformat('2023-04-03T00:00:00+00:00'),
            date=Date(2022, 1, 1),
            datetime=DateTime.fromisoformat('2022-01-01T10:00:00+00:00'),
            dct={'a': 1},
            firestore_id='my-id',
            lst=[1, 2, 3],
            obj=obj,
            updated_at=None
        )
        self.assertDictEqual(
            doc.to_dict(),
            {
                'byte': b'Text',
                'created_at': '2023-04-03T00:00:00+00:00',
                'date': '2022-01-01',
                'datetime': '2022-01-01T10:00:00+00:00',
                'dct': {'a': 1},
                'firestore_id': 'my-id',
                'lst': [1, 2, 3],
                'obj': compress(obj, serialize='pickle'),
                'updated_at': None
            }
        )
        Client.assert_not_called()

    def test_to_dict_excluded_keys(self, Client: mock.MagicMock):
        doc = firestore.Document(firestore_id='my-id', created_at='2023-04-03T00:00:00+00:00')
        self.assertDictEqual(
            doc.to_dict(),
            {
                'created_at': '2023-04-03T00:00:00+00:00',
                'firestore_id': 'my-id',
                'updated_at': None
            }
        )
        firestore.Document.config.excluded_keys = ['created_at', 'updated_at']
        self.assertDictEqual(doc.to_dict(), {'firestore_id': 'my-id'})
        firestore.Document.config.excluded_keys = []
        Client.assert_not_called()


@mock.patch.object(firestore.firestore, 'Client')
class LoadsPaginatedTestCase(TestCase):
    # pylint: disable=invalid-name

    @classmethod
    def setUpClass(cls) -> None:
        firestore.Document.config.collection_name = 'my-collection'

    @classmethod
    def tearDownClass(cls) -> None:
        firestore.Document.config.collection_name = None

    def setUp(self) -> None:
        firestore.FirestoreClient._connections = {}
        firestore.Document.config._client = None
        self.response = [
            firestore.Document(firestore_id='id01', type='1', mic='XYSX', created_at='2023-04-03T00:00:00+00:00'),
            firestore.Document(firestore_id='id02', type='2', mic='XZAG', created_at='2023-04-03T00:00:00+00:00'),
            firestore.Document(firestore_id='id03', type='3', mic='XZIM', created_at='2023-04-03T00:00:00+00:00'),
            firestore.Document(firestore_id='id04', type='4', mic='ZARX', created_at='2023-04-03T00:00:00+00:00')
        ]

    def test_loads_paginated_default(self, Client: mock.MagicMock):
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.get.return_value = self.response
        result = firestore.Document.loads_paginated()
        self.assertListEqual(result, self.response)
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.select.assert_not_called()
        Client.return_value.collection.return_value.order_by.assert_called_once_with('firestore_id')
        Client.return_value.collection.return_value.order_by.return_value.limit.assert_called_once_with(500)
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.get.assert_called_once_with()

    def test_loads_paginated_default_limit(self, Client: mock.MagicMock):
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.get.return_value = self.response
        result = firestore.Document.loads_paginated(limit=100)
        self.assertListEqual(result, self.response)
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.select.assert_not_called()
        Client.return_value.collection.return_value.order_by.assert_called_once_with('firestore_id')
        Client.return_value.collection.return_value.order_by.return_value.limit.assert_called_once_with(100)
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.get.assert_called_once_with()

    def test_loads_paginated_default_limit_more_query(self, Client: mock.MagicMock):
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.get.return_value = self.response
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.start_after.return_value.get.return_value = self.response[0:3]
        result = firestore.Document.loads_paginated(limit=4)
        self.assertListEqual(result, self.response + self.response[0:3])
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.select.assert_not_called()
        Client.return_value.collection.return_value.order_by.assert_called_once_with('firestore_id')
        Client.return_value.collection.return_value.order_by.return_value.limit.assert_called_once_with(4)
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.start_after.assert_called_once_with(self.response[-1])
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.get.assert_called_once_with()

    def test_loads_paginated_default_order_by(self, Client: mock.MagicMock):
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.get.return_value = self.response
        result = firestore.Document.loads_paginated(order_by='mic', limit=100)
        self.assertListEqual(result, self.response)
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.select.assert_not_called()
        Client.return_value.collection.return_value.order_by.assert_called_once_with('mic')
        Client.return_value.collection.return_value.order_by.return_value.limit.assert_called_once_with(100)
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.get.assert_called_once_with()

    def test_loads_paginated_default_order_by_error(self, Client: mock.MagicMock):
        Client.return_value.collection.return_value.select.return_value.order_by.return_value.limit.return_value.get.return_value = self.response
        self.assertRaisesRegex(
            ValueError,
            r"The order_by \(type\) must be in fields\(\['mic'\]\).",
            firestore.Document.loads_paginated,
            fields=['mic'],
            order_by='type',
            limit=100
        )
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.select.assert_not_called()
        Client.return_value.collection.return_value.select.return_value.order_by.assert_not_called()
        Client.return_value.collection.return_value.select.return_value.order_by.return_value.limit.assert_not_called()
        Client.return_value.collection.return_value.select.return_value.order_by.return_value.limit.return_value.get.assert_not_called()

    def test_loads_paginated_default_fields(self, Client: mock.MagicMock):
        new_response = [doc.fromkeys(['firestore_id', 'created_at', 'updated_at', 'mic']) for doc in self.response]
        Client.return_value.collection.return_value.select.return_value.order_by.return_value.limit.return_value.get.return_value = new_response
        result = firestore.Document.loads_paginated(fields=['mic'], order_by='mic', limit=100)
        self.assertListEqual(result, new_response)
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.select.assert_called_once_with(field_paths=['created_at', 'firestore_id', 'mic', 'updated_at'])
        Client.return_value.collection.return_value.select.return_value.order_by.assert_called_once_with('mic')
        Client.return_value.collection.return_value.select.return_value.order_by.return_value.limit.assert_called_once_with(100)
        Client.return_value.collection.return_value.select.return_value.order_by.return_value.limit.return_value.get.assert_called_once_with()


@mock.patch.object(redis_module.log, 'debug', mock.MagicMock)
class BaseDocumentCachedConfigTestCase(TestCase):

    def test_readonly_attributes(self):
        msg = "The field 'key_prefix' value cannot be changed."
        self.assertRaisesRegex(FieldValueError, msg, firestore.BaseDocumentCachedConfig, key_prefix='BANANA')
        with self.assertRaisesRegex(FieldValueError, msg):
            firestore.BaseDocumentCachedConfig(key_prefix='BANANA')

    def test_cache_property(self):
        obj1 = firestore.BaseDocumentCachedConfig()
        self.assertIsInstance(obj1.cache, firestore.RedisCacheCompressed)
        obj2 = firestore.BaseDocumentCachedConfig()
        self.assertNotEqual(obj1, obj2)

    def test_inheritance(self):
        self.assertEqual(
            firestore.BaseDocumentCachedConfig.__base__,
            firestore.BaseDocumentConfig
        )


@skip('Error on the test, need to fix it.')
@mock.patch.object(firestore.firestore, 'Client')
class DocumentCachedTestCase(TestCase):
    # pylint: disable=invalid-name

    @classmethod
    def setUpClass(cls) -> None:
        firestore.DocumentCached.config.collection_name = 'my-collection'
        cls.redis = redis_module.Redis(host=settings.REDIS_HOST, port=settings.REDIS_PORT)
        cls.key = '86085ac4d100a96d4e1739e87857bd4e266f6a2e4ad62b7ac5bee2aa62f88d2e'

    @classmethod
    def tearDownClass(cls) -> None:
        firestore.DocumentCached.config.collection_name = None

    def setUp(self) -> None:
        firestore.FirestoreClient._connections = {}
        firestore.DocumentCached.config._client = None
        self.redis.delete(self.key)

    def test_init(self, Client: mock.MagicMock):
        doc = firestore.DocumentCached(firestore_id='my-id', created_at='2023-04-03T00:00:00+00:00')
        self.assertDictEqual(
            doc,
            {
                'created_at': DateTime.fromisoformat('2023-04-03T00:00:00+00:00'),
                'firestore_id': 'my-id',
                'redis_key': self.key,
                'updated_at': None
            }
        )
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.document.assert_called_once_with('my-id')
        Client.return_value.collection.return_value.document.return_value.get.assert_called_once_with()

    def test_clear_cache_key(self, Client: mock.MagicMock):
        doc = firestore.DocumentCached(firestore_id='my-id')
        self.assertIsNotNone(doc.config.cache.get(doc.get_cache_key()))
        doc.clear_cache_key()
        self.assertIsNone(doc.config.cache.get(doc.get_cache_key()))
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.document.assert_called_once_with('my-id')
        Client.return_value.collection.return_value.document.return_value.get.assert_called_once_with()

    def test_get_cache_key(self, Client: mock.MagicMock):
        doc = firestore.DocumentCached(firestore_id='my-id')
        self.assertEqual(doc.get_cache_key(), 'firestore-document-redis-cached-my-collection-my-id')
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.document.assert_called_once_with('my-id')
        Client.return_value.collection.return_value.document.return_value.get.assert_called_once_with()

    def test_load_cache(self, Client: mock.MagicMock):
        # The first time that we load, if it is not on Redis we get it from Firestore and stores on Redis
        doc = firestore.DocumentCached(firestore_id='my-id', created_at='2023-11-11T00:00:00+00:00')
        # The second time that we load, it will get the info on Redis and update the object
        doc = firestore.DocumentCached(firestore_id='my-id', created_at='2023-04-03T00:00:00+00:00')
        self.assertDictEqual(
            doc,
            {
                'created_at': DateTime.fromisoformat('2023-11-11T00:00:00+00:00'),
                'firestore_id': 'my-id',
                'redis_key': '86085ac4d100a96d4e1739e87857bd4e266f6a2e4ad62b7ac5bee2aa62f88d2e',
                'updated_at': None
            }
        )
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.document.assert_called_once_with('my-id')
        Client.return_value.collection.return_value.document.return_value.get.assert_called_once_with()

    @mock.patch.object(firestore.DateTime, 'now', return_value='2023-11-11T00:00:00+00:00')
    def test_save(self, now: mock.MagicMock, Client: mock.MagicMock):
        doc = firestore.DocumentCached(firestore_id='my-id', created_at='2023-04-03T00:00:00+00:00')
        self.assertIsNotNone(doc.config.cache.get(doc.get_cache_key()))
        doc.save()
        self.assertIsNone(doc.config.cache.get(doc.get_cache_key()))
        Client.return_value.collection.return_value.document.return_value.set.assert_called_once_with(
            document_data={
                'redis_key': '86085ac4d100a96d4e1739e87857bd4e266f6a2e4ad62b7ac5bee2aa62f88d2e',
                'updated_at': '2023-11-11T00:00:00+00:00',
            },
            merge=True,
            timeout=60.0
        )
        now.assert_called_once_with()



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from unittest import TestCase, mock

from everysk.config import settings

from everysk.sdk.engines.market_data import MarketData

from everysk.core.datetime import Date
from everysk.core.http import HttpSDKPOSTConnection

###############################################################################
#   Market Data Test Case Implementation
###############################################################################
class MarketDataTestCase(TestCase):

    def setUp(self) -> None:
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()
        return super().setUp()

    ###############################################################################
    #  Get Security Test Case Implementation
    ###############################################################################
    def test_get_security_is_called(self):
        expected_data = '{"class_name": "MarketData", "method_name": "get_security", "self_obj": {"date": {"__date__": "2023-07-31"}, "__class_path__": "everysk.sdk.engines.market_data.MarketData"}, "params": {"date": "123", "ticker_list": ["test"], "ticker_type": ["choice1", "choice2"], "projection": "test_projection", "nearest": true}}'
        market_date = MarketData(date=Date(2023, 7, 31))

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            market_date.get_security(date='123', ticker_list=['test'], ticker_type=['choice1', 'choice2'], projection='test_projection', nearest=True)

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )

    ###############################################################################
    #  Get Historical Test Case Implementation
    ###############################################################################
    def test_get_historical_is_called(self):
        expected_data = '{"class_name": "MarketData", "method_name": "get_historical", "self_obj": {"date": {"__date__": "2023-07-31"}, "__class_path__": "everysk.sdk.engines.market_data.MarketData"}, "params": {"date": "2023-08-14", "start_date": "2022-05-14", "end_date": "2024-07-17", "ticker_list": ["test"], "ticker_type": ["choice1", "choice2"], "projection": "test_projection", "nearest": true, "real_time": true}}'
        market_date = MarketData(date=Date(2023, 7, 31))

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            market_date.get_historical(date='2023-08-14', start_date='2022-05-14', end_date='2024-07-17', ticker_list=['test'], ticker_type=['choice1', 'choice2'], projection='test_projection', nearest=True, real_time=True)

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
import unittest
from unittest import mock

from everysk.config import settings
from everysk.core.datetime import DateTime
from everysk.core.exceptions import SDKValueError
from everysk.core.http import HttpSDKPOSTConnection

from everysk.sdk.entities.worker_template.base import WorkerTemplate
from everysk.sdk.entities.query import Query

###############################################################################
#   Report TestCase Implementation
###############################################################################
class WorkerTemplateTestCase(unittest.TestCase):

    def setUp(self):
        self.sample_data = {
            'created_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'updated_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'version': 'v1',
            'id': 'wrkt_xyz',
            'sort_index': 1,
            'category': 'portfolio',
            'form_outputs': {
                'OUTPUT_ONE': [
                    {
                        'key': 'value',
                    }
                ]
            },
            'script_entry_point': 'main',
            'tags': [],
            'type': 'BASIC',
            'script_visible_source': "def main(args):\n    print('test')\n    return",
            'form_functions': {},
            'ports': [
                {
                    'key': 'value',

                }
            ],
            'name': 'Worker Teste',
            'script_source': "def main(args):\n    print('test')\n    return",
            'icon': 'retriever',
            'script_runtime': 'python',
            'description': '#Worker Sample',
            'form_inputs': [
                {
                    'key': 'value'
                }

            ],
            'visible': True,
            'default_output': 'OUTPUT_ONE',
        }
        self.worker_template = WorkerTemplate(**self.sample_data)
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()

    def test_get_id_prefix(self):
        self.assertEqual(WorkerTemplate.get_id_prefix(), settings.WORKER_TEMPLATE_ID_PREFIX)

    def test_validate(self):
        expected_data = '{"class_name": "WorkerTemplate", "method_name": "validate", "self_obj": {"path_name": null, "created_on": {"__datetime__": "2023-09-09T09:09:09.000009+00:00"}, "updated_on": {"__datetime__": "2023-09-09T09:09:09.000009+00:00"}, "version": "v1", "id": "wrkt_xyz", "sort_index": 1, "category": "portfolio", "form_outputs": {"OUTPUT_ONE": [{"key": "value"}]}, "script_entry_point": "main", "tags": [], "type": "BASIC", "script_visible_source": "def main(args):\\n    print(\'test\')\\n    return", "form_functions": {}, "ports": [{"key": "value"}], "name": "Worker Teste", "script_source": "def main(args):\\n    print(\'test\')\\n    return", "icon": "retriever", "script_runtime": "python", "description": "#Worker Sample", "form_inputs": [{"key": "value"}], "visible": true, "default_output": "OUTPUT_ONE", "__class_path__": "everysk.sdk.entities.worker_template.base.WorkerTemplate"}, "params": {}}'
        worker_template: WorkerTemplate = self.worker_template.copy()

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            worker_template.validate()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )

    ###############################################################################
    #   Check Query Test Case Implementation
    ###############################################################################
    def test_check_query_returns_valid(self):
        worker_template: WorkerTemplate = self.worker_template
        result = worker_template._check_query(query='test_query') # pylint: disable=protected-access

        self.assertEqual(result, True)

    def test_check_entity_to_query_raises_sdk_error(self):
        with self.assertRaises(SDKValueError) as context:
            worker_template: WorkerTemplate = self.worker_template
            worker_template['tags'] = ['tag1', 'tag2']
            worker_template._check_entity_to_query() # pylint: disable=protected-access

        self.assertEqual(str(context.exception), "Can't filter by Name and Tags at the same time")

    ###############################################################################
    #   Mount Query Test Case Implementation
    ###############################################################################
    def test_mount_query_with_name(self):
        expected_query = Query(**{'_klass': WorkerTemplate, 'filters': [('name', '<', 'worker testf'), ('name', '>=', 'worker teste')], 'order': [], 'projection': None, 'distinct_on': [], 'limit': 1, 'offset': None, 'page_size': None, 'page_token': None})
        worker_template: WorkerTemplate = self.worker_template
        result = worker_template.to_query(limit=1)

        self.assertEqual(result, expected_query)

    def test_mount_query_with_tags(self):
        expected_query = Query(**{'_klass': WorkerTemplate, 'filters': [('tags', '=', 'tag1'), ('tags', '=', 'tag2')], 'order': [], 'projection': None, 'distinct_on': [], 'limit': 1, 'offset': None, 'page_size': None, 'page_token': None})
        worker_template: WorkerTemplate = self.worker_template
        worker_template['name'] = None
        worker_template['tags'] = ['tag1', 'tag2']
        result = worker_template.to_query(limit=1)

        self.assertEqual(result, expected_query)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
import unittest
from unittest import mock
from everysk.config import settings

from everysk.core.datetime import DateTime, ZoneInfo
from everysk.core.exceptions import SDKValueError
from everysk.core.http import HttpSDKPOSTConnection

from everysk.sdk.entities.report.base import Report

###############################################################################
#   Report TestCase Implementation
###############################################################################
class ReportTestCase(unittest.TestCase):

    def setUp(self):
        self.sample_data = {
            'created_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'updated_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'version': 'v1',
            'id': 'repo_12345678',
            'name': 'My Report',
            'description': 'This is a sample report.',
            'tags': ['tag1',
            'tag2'],
            'link_uid': None,
            'workspace': 'my_workspace',
            'date': DateTime(2022, 1, 1, 12, 0, tzinfo=ZoneInfo(key='UTC')),
            'widgets': [{'type': 'chart', 'data': {}}],
            'url': '/sefdsf5s54sdfsksdfs5',
            'authorization': 'private',
            'config_cascaded': {'setting1': 'value1',
            'setting2': 'value2'},
            'layout_content': {'section1': {}}
        }
        self.report = Report(**self.sample_data)
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()

    def test_get_id_prefix(self):
        self.assertEqual(Report.get_id_prefix(), settings.REPORT_ID_PREFIX)

    def test_validate(self):
        expected_data = '{"class_name": "Report", "method_name": "validate", "self_obj": {"created_on": {"__datetime__": "2023-09-09T09:09:09.000009+00:00"}, "updated_on": {"__datetime__": "2023-09-09T09:09:09.000009+00:00"}, "version": "v1", "id": "repo_12345678", "name": "My Report", "description": "This is a sample report.", "tags": ["tag1", "tag2"], "link_uid": null, "workspace": "my_workspace", "date": "20220101", "widgets": [{"type": "chart", "data": {}}], "url": "/sefdsf5s54sdfsksdfs5", "authorization": "private", "config_cascaded": {"setting1": "value1", "setting2": "value2"}, "layout_content": {"section1": {}}, "__class_path__": "everysk.sdk.entities.report.base.Report"}, "params": {}}'
        report: Report = self.report.copy()

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            report.validate()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )

    def test_query_load_with_id(self):
        expected_data = '{"class_name": "Report", "method_name": "retrieve", "self_obj": null, "params": {"entity_id": "repo_1234567891011211234567890"}}'

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            Report(id='repo_1234567891011211234567890', workspace='SampleWorkspace').load()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )

    def test_query_load(self):
        expected_data = '{"class_name": "Query", "method_name": "load", "self_obj": {"_klass": "Report", "filters": [["workspace", "=", "SampleWorkspace"], ["link_uid", "=", "SampleLinkUID"]], "order": [], "projection": null, "distinct_on": [], "limit": null, "offset": null, "page_size": null, "page_token": null, "_clean_order": [], "__class_path__": "everysk.sdk.entities.query.Query"}, "params": {"offset": null}}'

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            Report(link_uid='SampleLinkUID', workspace='SampleWorkspace').load()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )

    def test_entity_to_query(self):
        cmp_query = Report.query.where('url', self.report.url)
        query = Report(url=self.report.url).to_query() #pylint: disable=protected-access

        self.assertDictEqual(query, cmp_query)

    def test_check_query_raises_sdk_error(self):
        with self.assertRaises(SDKValueError) as context:
            Report(url=self.report.url, name='report_name').to_query()

        self.assertEqual(str(context.exception), "Can't filter by URL and Name, Tags or Link UID at the same time")
